import { Cached, CacheTimestampInterface, Redis } from '@gapizza/redis';
import admin from 'firebase-admin';
import { DeepPartial } from 'ts-essentials';
import { firestore } from 'firebase-admin/lib/firestore';
import Timestamp = firestore.Timestamp;
export declare enum FILTER_OPERATORS {
    GT = ">",
    GTE = ">=",
    LT = "<",
    LTE = "<=",
    EQ = "==",
    NOT_EQ = "!=",
    IN = "in",
    NOT_IN = "not-in",
    CONTAINS = "array-contains",
    CONTAINS_ANY = "array-contains-any"
}
export declare type DalModelValue = string | Array<string> | Date | number | null | boolean;
export interface ListFilterInterface {
    property: string;
    operator: FILTER_OPERATORS;
    value: DalModelValue;
}
export declare enum SORT_DIRECTION {
    ASC = "asc",
    DESC = "desc"
}
export interface ListSortInterface {
    property: string;
    direction: SORT_DIRECTION;
}
export interface QueryInterface {
    filters?: ListFilterInterface[];
    sort?: ListSortInterface;
    offset?: number;
    limit?: number;
    before?: DalModelValue;
    after?: DalModelValue;
}
export interface DalModel {
    id: string;
    validate(): void | Promise<void>;
    createdAt: Date;
    updatedAt: Date;
}
interface ServicesInterface {
    firestore: admin.firestore.Firestore;
    redis: Redis;
}
export interface FirestoreConfigInterface<T extends DalModel> {
    readTimestampsToDates?: boolean;
    collection: string;
    convertForDb(instance: DeepPartial<T>): any;
    convertFromDb(params: any): T | Promise<T>;
}
export interface FirestoreCacheConfigInterface<T extends DalModel> {
    cacheTtlSec: number;
    stringifyForCache(instance: T): Promise<string> | string;
    parseFromCache(instance: string): Promise<T> | T;
}
/**
 * @class
 */
export declare class Firestore<T extends DalModel> extends Cached<T> {
    /**
     * @param {ServicesInterface} services
     */
    constructor(services: ServicesInterface);
    /**
     * Translate dates to timestamp
     *
     * @param {any} obj
     * @returns {any}
     */
    static translateDatesToTimestamps(obj: any): any;
    /**
     * Translate timestamps to dates
     *
     * @param {any} obj
     * @returns {any}
     */
    static translateTimestampsToDates(obj: any): any;
    /**
     * Get cache timestamp from firestore timestamp, or fall back to redisTimestamp
     *
     * @param {FirebaseFirestore.Timestamp} timestamp
     * @returns {CacheTimestampInterface}
     */
    static getCacheTimestamp(timestamp: Timestamp): CacheTimestampInterface;
    /**
     * Configure firestore
     *
     * @param {FirestoreConfigInterface<T>} config
     * @param {FirestoreCacheConfigInterface<T>} cacheConfig
     * @returns {void}
     */
    configure(config: FirestoreConfigInterface<T>, cacheConfig?: FirestoreCacheConfigInterface<T>): void;
    readonly services: ServicesInterface;
    private config?;
    /**
     * Clean model of common properties that shouldn't be written
     *
     * @param {{}} model
     * @returns {{}}
     */
    private static cleanModel;
    /**
     * Build firestore query from structured query
     * NOTE: Firestore doesn't work as expected when you combine endBefore and limit, and this corrects that
     *
     * @param {QueryInterface} query
     * @returns {Query}
     */
    private getQuerySnapshot;
    /**
     * Create instance of model in db
     *
     * @param {T} instance
     * @returns {Promise<T>}
     */
    create(instance: T): Promise<T>;
    /**
     * Get instance
     *
     * @param {string} id
     * @returns {Promise<T | null>}
     */
    get(id: string): Promise<T | null>;
    /**
     * Get instance without touching cache
     *
     * @param {string} id
     * @returns {Promise<T | null>}
     */
    internalGet(id: string): Promise<{
        instance: T | null;
        timestamp: CacheTimestampInterface;
    }>;
    /**
     * Get value directly from the db, by-passing cache and convertFromDb
     *
     * @param {string} id
     * @returns {Promise<any | null>}
     */
    rawGet(id: string): Promise<any | null>;
    /**
     * Get value directly from the db, by-passing cache and convertFromDb
     *
     * @param {string} id
     * @returns {Promise<{ instance: any, timestamp: CacheTimestampInterface }>}
     */
    private internalRawGet;
    /**
     * Get instance or throw
     *
     * @param {string} id
     * @param {boolean} throw404
     * @returns {Promise<T>}
     */
    getOrThrow(id: string, throw404?: boolean): Promise<T>;
    /**
     * Internal get or throw without touching cache
     *
     * @param {string} id
     * @param {boolean} throw404
     * @returns {Promise<T>}
     */
    private internalGetOrThrow;
    /**
     * Update properties of model
     *
     * @param {string} id
     * @param {{}} patchUpdate
     * @param {Date} [curDate]
     * @returns {Promise<T>}
     */
    patch(id: string, patchUpdate: DeepPartial<T>, curDate?: Date): Promise<T>;
    /**
     * Model exists
     *
     * @param {string} id
     * @returns {Promise<boolean>}
     */
    exists(id: string): Promise<boolean>;
    /**
     * Remove model
     *
     * @param {string} id
     * @returns {Promise<void>}
     */
    remove(id: string): Promise<void>;
    /**
     * Remove model without touching cache
     *
     * @param {string} id
     * @returns {Promise<void>}
     */
    private internalRemove;
    /**
     * Overwrite model
     *
     * @param {string} id
     * @param {{}} instance
     * @param {Date} [curDate]
     * @returns {Promise<T>}
     */
    update(id: string, instance: T, curDate?: Date): Promise<T>;
    /**
     * List models satisfying query
     *
     * @param {QueryInterface} query
     * @returns {Promise<T[]>}
     */
    query(query?: QueryInterface): Promise<T[]>;
    /**
     * Get value directly from the db, by-passing cache and convertFromDb
     *
     * @param {QueryInterface} query
     * @returns {Promise<any[]>}
     */
    rawQuery(query: QueryInterface): Promise<any[]>;
    /**
     * Remove by query
     *
     * @param {QueryInterface} query
     * @param paginate
     * @param {PAGINATE_CONCURRENCY} pageSize
     * @returns {Promise<void>}
     */
    removeByQuery(query: QueryInterface, paginate?: boolean, pageSize?: number): Promise<string[]>;
}
export {};
