"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Firestore = exports.SORT_DIRECTION = exports.FILTER_OPERATORS = void 0;
const redis_1 = require("@gapizza/redis");
const bluebird_1 = __importDefault(require("bluebird"));
const class_transformer_1 = require("class-transformer");
const clean_deep_1 = __importDefault(require("clean-deep"));
const err_1 = __importDefault(require("err"));
const fast_json_stable_stringify_1 = __importDefault(require("fast-json-stable-stringify"));
const firebase_admin_1 = __importDefault(require("firebase-admin"));
const flat_1 = __importDefault(require("flat"));
const http_status_1 = __importDefault(require("http-status"));
const lodash_1 = require("lodash");
const luxon_1 = require("luxon");
const traverse_1 = __importDefault(require("traverse"));
const logger_1 = __importDefault(require("./logger"));
var FILTER_OPERATORS;
(function (FILTER_OPERATORS) {
    FILTER_OPERATORS["GT"] = ">";
    FILTER_OPERATORS["GTE"] = ">=";
    FILTER_OPERATORS["LT"] = "<";
    FILTER_OPERATORS["LTE"] = "<=";
    FILTER_OPERATORS["EQ"] = "==";
    FILTER_OPERATORS["NOT_EQ"] = "!=";
    FILTER_OPERATORS["IN"] = "in";
    FILTER_OPERATORS["NOT_IN"] = "not-in";
    FILTER_OPERATORS["CONTAINS"] = "array-contains";
    FILTER_OPERATORS["CONTAINS_ANY"] = "array-contains-any";
})(FILTER_OPERATORS = exports.FILTER_OPERATORS || (exports.FILTER_OPERATORS = {}));
var SORT_DIRECTION;
(function (SORT_DIRECTION) {
    SORT_DIRECTION["ASC"] = "asc";
    SORT_DIRECTION["DESC"] = "desc";
})(SORT_DIRECTION = exports.SORT_DIRECTION || (exports.SORT_DIRECTION = {}));
const CLEAN_CONFIG = {
    emptyArrays: false,
    emptyObjects: false,
    emptyStrings: false,
    nullValues: false,
    undefinedValues: true,
};
const CONFIG_ERROR = 'firestore instance not configured';
const PAGINATE_CONCURRENCY = 25;
/**
 * @class
 */
class Firestore extends redis_1.Cached {
    /**
     * @param {ServicesInterface} services
     */
    constructor(services) {
        super();
        this.services = services;
    }
    /* eslint-disable @typescript-eslint/explicit-module-boundary-types */
    /**
     * Translate dates to timestamp
     *
     * @param {any} obj
     * @returns {any}
     */
    static translateDatesToTimestamps(obj) {
        // eslint-disable-next-line array-callback-return,func-names
        return (0, traverse_1.default)((0, class_transformer_1.classToPlain)(obj)).map(function (property) {
            if ((0, lodash_1.isDate)(property)) {
                this.update(firebase_admin_1.default.firestore.Timestamp.fromDate(property));
            }
        });
    }
    /**
     * Translate timestamps to dates
     *
     * @param {any} obj
     * @returns {any}
     */
    static translateTimestampsToDates(obj) {
        // eslint-disable-next-line array-callback-return,func-names
        return (0, traverse_1.default)(obj).map(function (property) {
            if (property instanceof firebase_admin_1.default.firestore.Timestamp) {
                this.update(property.toDate());
            }
        });
    }
    /* eslint-enable @typescript-eslint/explicit-module-boundary-types */
    /**
     * Get cache timestamp from firestore timestamp, or fall back to redisTimestamp
     *
     * @param {FirebaseFirestore.Timestamp} timestamp
     * @returns {CacheTimestampInterface}
     */
    static getCacheTimestamp(timestamp) {
        return {
            seconds: timestamp.seconds,
            // eslint-disable-next-line no-magic-numbers
            microseconds: (0, lodash_1.round)(timestamp.nanoseconds / 1000),
        };
    }
    /**
     * Configure firestore
     *
     * @param {FirestoreConfigInterface<T>} config
     * @param {FirestoreCacheConfigInterface<T>} cacheConfig
     * @returns {void}
     */
    configure(config, cacheConfig) {
        this.config = { readTimestampsToDates: false, ...config };
        const { redis } = this.services;
        if (cacheConfig) {
            const { cacheTtlSec: ttlSec, stringifyForCache, parseFromCache } = cacheConfig;
            this.configureCache({ redis }, { ttlSec, stringifyForCache, parseFromCache, prefix: config.collection });
        }
    }
    /**
     * Clean model of common properties that shouldn't be written
     *
     * @param {{}} model
     * @returns {{}}
     */
    static cleanModel(model) {
        model = (0, class_transformer_1.classToPlain)(model);
        if (model.createdAt)
            delete model.createdAt;
        if (model.id)
            delete model.id;
        return (0, clean_deep_1.default)(model, CLEAN_CONFIG);
    }
    /* eslint-disable sonarjs/cognitive-complexity */
    /**
     * Build firestore query from structured query
     * NOTE: Firestore doesn't work as expected when you combine endBefore and limit, and this corrects that
     *
     * @param {QueryInterface} query
     * @returns {Query}
     */
    async getQuerySnapshot(query) {
        if (!this.config)
            throw new err_1.default(CONFIG_ERROR);
        let reverse = false;
        query = Firestore.translateDatesToTimestamps(query);
        let ref = this.services.firestore.collection(this.config.collection);
        if (query.filters) {
            ref = (0, lodash_1.reduce)(query.filters, (result, filter) => result.where(filter.property, filter.operator, filter.value), ref);
        }
        if (query.offset) {
            ref = ref.offset(query.offset);
        }
        if (query.sort) {
            if (query.before && query.limit) {
                ref = ref.orderBy(query.sort.property, query.sort.direction === SORT_DIRECTION.DESC ? SORT_DIRECTION.ASC : SORT_DIRECTION.DESC);
                reverse = true;
            }
            else {
                ref = ref.orderBy(query.sort.property, query.sort.direction);
            }
        }
        if (!(0, lodash_1.isNil)(query.before) && !(0, lodash_1.isNil)(query.after)) {
            throw new err_1.default('cannot provide both before and after for pagination');
        }
        if ((!(0, lodash_1.isNil)(query.before) || !(0, lodash_1.isNil)(query.after)) && !query.sort) {
            throw new err_1.default('if before or after is provided, must provide sort');
        }
        if (query.before) {
            ref = reverse ? ref.startAfter(query.before) : ref.endBefore(query.before);
        }
        if (query.after) {
            ref = ref.startAfter(query.after);
        }
        if (query.limit) {
            ref = ref.limit(query.limit);
        }
        return {
            reverse,
            querySnapshot: await ref.get(),
        };
    }
    /* eslint-enable sonarjs/cognitive-complexity */
    /**
     * Create instance of model in db
     *
     * @param {T} instance
     * @returns {Promise<T>}
     */
    async create(instance) {
        if (!this.config)
            throw new err_1.default(CONFIG_ERROR);
        await instance.validate();
        // I don't know why that casting is necessary
        const data = await this.config.convertForDb(instance);
        if (!(0, lodash_1.isDate)(data.createdAt)) {
            throw new err_1.default('createdAt must be a Date');
        }
        if (!(0, lodash_1.isDate)(data.updatedAt)) {
            throw new err_1.default('updatedAt must be a Date');
        }
        try {
            const cleanedData = (0, clean_deep_1.default)(Firestore.translateDatesToTimestamps(data), CLEAN_CONFIG);
            const { writeTime } = await this.services.firestore.collection(this.config.collection).doc(instance.id).create(cleanedData);
            await this.cache.delLists();
            await this.cache.setSafe(instance.id, instance, Firestore.getCacheTimestamp(writeTime));
            return instance;
        }
        catch (error) {
            // eslint-disable-next-line no-magic-numbers
            if ((error === null || error === void 0 ? void 0 : error.code) === 6) {
                throw new err_1.default('Already exists', http_status_1.default.CONFLICT);
            }
            else {
                logger_1.default.error(`Error during create: ${error === null || error === void 0 ? void 0 : error.message}`);
                throw new err_1.default('Could not create');
            }
        }
    }
    /**
     * Get instance
     *
     * @param {string} id
     * @returns {Promise<T | null>}
     */
    async get(id) {
        const cached = await this.cache.get(id);
        if (cached)
            return cached;
        const { instance, timestamp } = await this.internalGet(id);
        await (instance ? this.cache.setSafe(id, instance, timestamp) : this.cache.delSafe(id, timestamp));
        return instance;
    }
    /**
     * Get instance without touching cache
     *
     * @param {string} id
     * @returns {Promise<T | null>}
     */
    async internalGet(id) {
        if (!this.config)
            throw new err_1.default(CONFIG_ERROR);
        const snapshot = await this.services.firestore.collection(this.config.collection).doc(id).get();
        const timestamp = Firestore.getCacheTimestamp(snapshot.updateTime || snapshot.createTime || snapshot.readTime);
        if (!snapshot.exists) {
            return {
                instance: null,
                timestamp,
            };
        }
        const data = this.config.readTimestampsToDates ? Firestore.translateTimestampsToDates(snapshot.data()) : snapshot.data();
        try {
            const result = data ? await this.config.convertFromDb({ id, ...data }) : null;
            return {
                instance: result,
                timestamp,
            };
        }
        catch (error) {
            logger_1.default.error(`Error while reading from db: ${error === null || error === void 0 ? void 0 : error.message}`);
            logger_1.default.error('Data: ', data);
            throw error;
        }
    }
    /**
     * Get value directly from the db, by-passing cache and convertFromDb
     *
     * @param {string} id
     * @returns {Promise<any | null>}
     */
    async rawGet(id) {
        const { raw } = await this.internalRawGet(id);
        return raw;
    }
    /**
     * Get value directly from the db, by-passing cache and convertFromDb
     *
     * @param {string} id
     * @returns {Promise<{ instance: any, timestamp: CacheTimestampInterface }>}
     */
    async internalRawGet(id) {
        if (!this.config)
            throw new err_1.default(CONFIG_ERROR);
        const snapshot = await this.services.firestore.collection(this.config.collection).doc(id).get();
        const timestamp = Firestore.getCacheTimestamp(snapshot.updateTime || snapshot.createTime || snapshot.readTime);
        if (!snapshot.exists)
            return { raw: null, timestamp };
        const data = this.config.readTimestampsToDates ? Firestore.translateTimestampsToDates(snapshot.data()) : snapshot.data();
        return {
            raw: data ? { id, ...data } : null,
            timestamp,
        };
    }
    /**
     * Get instance or throw
     *
     * @param {string} id
     * @param {boolean} throw404
     * @returns {Promise<T>}
     */
    async getOrThrow(id, throw404 = false) {
        const cached = await this.cache.get(id);
        if (cached)
            return cached;
        const { instance, timestamp } = await this.internalGetOrThrow(id, throw404);
        await this.cache.setSafe(id, instance, timestamp);
        return instance;
    }
    /**
     * Internal get or throw without touching cache
     *
     * @param {string} id
     * @param {boolean} throw404
     * @returns {Promise<T>}
     */
    async internalGetOrThrow(id, throw404 = false) {
        const { instance, timestamp } = await this.internalGet(id);
        if (!instance)
            throw new err_1.default(`id: ${id} not found`, throw404 ? http_status_1.default.NOT_FOUND : http_status_1.default.INTERNAL_SERVER_ERROR);
        return { instance, timestamp };
    }
    /**
     * Update properties of model
     *
     * @param {string} id
     * @param {{}} patchUpdate
     * @param {Date} [curDate]
     * @returns {Promise<T>}
     */
    async patch(id, patchUpdate, curDate = luxon_1.DateTime.utc().toJSDate()) {
        if (!this.config)
            throw new err_1.default(CONFIG_ERROR);
        const flattened = Firestore.cleanModel((0, flat_1.default)({ ...(await this.config.convertForDb(patchUpdate)), updatedAt: curDate }, { safe: true }));
        await this.services.firestore
            .collection(this.config.collection)
            .doc(id)
            .update(Firestore.translateDatesToTimestamps(flattened));
        const { raw, timestamp } = await this.internalRawGet(id);
        const instance = await this.config.convertFromDb(raw);
        await this.cache.setSafe(id, instance, timestamp);
        await this.cache.delLists();
        return instance;
    }
    /**
     * Model exists
     *
     * @param {string} id
     * @returns {Promise<boolean>}
     */
    async exists(id) {
        return !!(await this.get(id));
    }
    /**
     * Remove model
     *
     * @param {string} id
     * @returns {Promise<void>}
     */
    async remove(id) {
        const timestamp = await this.internalRemove(id);
        await this.cache.delSafe(id, timestamp);
        await this.cache.delLists();
    }
    /**
     * Remove model without touching cache
     *
     * @param {string} id
     * @returns {Promise<void>}
     */
    async internalRemove(id) {
        if (!this.config)
            throw new err_1.default(CONFIG_ERROR);
        const { writeTime } = await this.services.firestore.collection(this.config.collection).doc(id).delete();
        return Firestore.getCacheTimestamp(writeTime);
    }
    /**
     * Overwrite model
     *
     * @param {string} id
     * @param {{}} instance
     * @param {Date} [curDate]
     * @returns {Promise<T>}
     */
    async update(id, instance, curDate = luxon_1.DateTime.utc().toJSDate()) {
        if (!this.config)
            throw new err_1.default(CONFIG_ERROR);
        await instance.validate();
        // Retain the original createdAt, and ensure it exists
        const { createdAt } = await this.getOrThrow(id);
        // I don't know why that casting is necessary
        const updated = { ...Firestore.cleanModel({ ...(await this.config.convertForDb(instance)), updatedAt: curDate }), createdAt };
        const { writeTime } = await this.services.firestore.collection(this.config.collection).doc(id).set(Firestore.translateDatesToTimestamps(updated));
        const timestamp = Firestore.getCacheTimestamp(writeTime);
        const updatedInstance = await this.config.convertFromDb({ id, ...updated });
        await this.cache.setSafe(id, updatedInstance, timestamp);
        await this.cache.delLists();
        return updatedInstance;
    }
    /**
     * List models satisfying query
     *
     * @param {QueryInterface} query
     * @returns {Promise<T[]>}
     */
    async query(query = {}) {
        const cacheKey = (0, fast_json_stable_stringify_1.default)(query);
        const cacheResults = await this.cache.getList(cacheKey);
        if (cacheResults)
            return cacheResults;
        const { reverse, querySnapshot } = await this.getQuerySnapshot(query);
        const snapshots = [];
        querySnapshot.forEach((snapshot) => {
            if (!snapshot.exists)
                return;
            if (!this.config)
                throw new err_1.default(CONFIG_ERROR);
            const data = this.config.readTimestampsToDates ? Firestore.translateTimestampsToDates(snapshot.data()) : snapshot.data();
            snapshots.push({ id: snapshot.id, ...data });
        });
        // eslint-disable-next-line unicorn/no-array-method-this-argument
        const results = await bluebird_1.default.map(snapshots, (snapshot) => {
            if (!this.config)
                throw new err_1.default(CONFIG_ERROR);
            return this.config.convertFromDb(snapshot);
        });
        if (reverse) {
            results.reverse();
        }
        const timestamp = Firestore.getCacheTimestamp(querySnapshot.readTime);
        await this.cache.setListSafe(cacheKey, results, timestamp);
        return results;
    }
    /**
     * Get value directly from the db, by-passing cache and convertFromDb
     *
     * @param {QueryInterface} query
     * @returns {Promise<any[]>}
     */
    async rawQuery(query) {
        const { reverse, querySnapshot } = await this.getQuerySnapshot(query);
        const results = [];
        querySnapshot.forEach((snapshot) => {
            if (!snapshot.exists)
                return;
            if (!this.config)
                throw new err_1.default(CONFIG_ERROR);
            const data = this.config.readTimestampsToDates ? Firestore.translateTimestampsToDates(snapshot.data()) : snapshot.data();
            results.push({ id: snapshot.id, ...data });
        });
        if (reverse) {
            results.reverse();
        }
        return results;
    }
    /**
     * Remove by query
     *
     * @param {QueryInterface} query
     * @param paginate
     * @param {PAGINATE_CONCURRENCY} pageSize
     * @returns {Promise<void>}
     */
    async removeByQuery(query, paginate = false, pageSize = PAGINATE_CONCURRENCY) {
        let ids = [];
        let isReverse = false;
        const internalRemoveByQuery = async (_query) => {
            const { reverse, querySnapshot } = await this.getQuerySnapshot(_query);
            const pageIds = [];
            querySnapshot.forEach((snapshot) => {
                if (!snapshot.exists)
                    return;
                pageIds.push(snapshot.id);
            });
            await bluebird_1.default.each(pageIds, async (id) => {
                const timestamp = await this.internalRemove(id);
                await this.cache.delSafe(id, timestamp);
            });
            ids = [...ids, ...pageIds];
            isReverse = reverse;
            return pageIds;
        };
        if (paginate) {
            const nextPage = async (after) => {
                const paginatedQuery = after
                    ? { sort: { property: 'id', direction: SORT_DIRECTION.ASC }, ...query, limit: pageSize, after }
                    : { sort: { property: 'id', direction: SORT_DIRECTION.ASC }, ...query, limit: pageSize };
                const pageIds = await internalRemoveByQuery(paginatedQuery);
                if (pageIds.length >= pageSize) {
                    await nextPage((0, lodash_1.last)(pageIds));
                }
            };
            await nextPage();
        }
        else {
            await internalRemoveByQuery(query);
        }
        if (isReverse) {
            ids.reverse();
        }
        await this.cache.delLists();
        return ids;
    }
}
exports.Firestore = Firestore;
//# sourceMappingURL=firestore.js.map